import numpy as np
import pandas as pd
import os
import pickle
import anndata as ad
import multiprocessing 
import time
import pickle
import warnings
import random



warnings.simplefilter(action='ignore', category=FutureWarning)
warnings.filterwarnings("ignore", category=Warning)


class get_CTSS_count():


    def __init__(self,out_dir,minCTSSCount,minFC,n_proc,windowSize):
        self.out_dir=out_dir
        generefPath=os.path.join(out_dir,'ref_file','ref_gene.tsv')
        fetched_reads_path=os.path.join(out_dir,'count','fetch_reads.pkl')
        afterfiltered_path=os.path.join(out_dir,'count','afterfiltered.csv')

        self.generefdf=pd.read_csv(generefPath,delimiter='\t')
        with open(fetched_reads_path,'rb') as f:
            self.fetchadata=pickle.load(f)
        self.alloneclusterdf=pd.read_csv(afterfiltered_path)

        self.minCTSSCount=minCTSSCount
        self.minFC=minFC
        self.n_proc=n_proc
        self.windowSize=windowSize



    def window_sliding(self,genereads,TSS_start,TSS_end,strand):


        leftIndex=0
        filterls=[i for i in genereads if ('14S' in i[2]) or ('15S' in i[2]) or ('16S' in i[2])  ]
        #print("the pid of current process：",os.getpid())
        
        #calculate the TSS position and corresponding counts
        promoterTSS = [read[0] for read in filterls if (read[0] >= TSS_start) and (read[0] <= TSS_end)]
        TSS,count=np.unique(promoterTSS,return_counts=True)



        originaldf=pd.DataFrame({'count':count},index=TSS)
        desireddf=pd.DataFrame({'count':0},index=[i for i in range(int(TSS_start),int(TSS_end))])
        desireddf['count']=desireddf.index.map(originaldf['count']).fillna(0)
        desireddf.reset_index(inplace=True)
        nonzeroarray=np.array(desireddf)


        if strand=='+':
            sortfinalarray=nonzeroarray[nonzeroarray[:, 0].argsort()]
        elif strand=='-':
            sortfinalarray=nonzeroarray[nonzeroarray[:, 0].argsort()[::-1]]
        
        TSS=sortfinalarray.T[0]
        count=sortfinalarray.T[1]


        #do something with sliding windows algorithm   
        storels=[]
        for i in range(len(TSS) - self.windowSize + 1):
            #print(i)
            onewindow=TSS[i: i + self.windowSize]
            correspondingcount=count[i: i + self.windowSize]
            middlecount=correspondingcount[leftIndex]
            foldchange=(middlecount+1)/(sum(correspondingcount)/len(correspondingcount)+1)
            storels.append([onewindow[leftIndex],correspondingcount[leftIndex],foldchange])

            
        foldchangels=[i[2] for i in storels]
        sortindex=sorted(range(len(foldchangels)), key=lambda k: foldchangels[k],reverse=True)
        allsortls=[storels[i] for i in sortindex]

        return allsortls




    def _get_CTSS(self):
        alloneclusterdf=self.alloneclusterdf
        alloneclusterdf['gene_id']=alloneclusterdf['Unnamed: 0'].str.split('*',expand=True)[0]
        alloneclusterdf['TSS_start']=alloneclusterdf['Unnamed: 0'].str.split('*',expand=True)[1].str.split('_',expand=True)[0].astype('float')
        alloneclusterdf['TSS_end']=alloneclusterdf['Unnamed: 0'].str.split('_',expand=True)[1].astype('float')

        self.generefdf.reset_index(inplace=True)
        

        #print(self.generefdf)
        stranddf=self.generefdf[['Strand','gene_id']]
        alloneclusterdf=alloneclusterdf.merge(stranddf,on='gene_id')


        start_time=time.time()

        #pool = multiprocessing.Pool(processes=self.n_proc)
        allsortfddict={}

        for i in range(0,len(alloneclusterdf)):
            geneID=alloneclusterdf['gene_id'][i]
            #print(geneID)
            genereads=self.fetchadata[geneID]
            clusterID=alloneclusterdf['Unnamed: 0'][i]
            TSS_start=alloneclusterdf['TSS_start'][i]
            TSS_end=alloneclusterdf['TSS_end'][i]
            strand=alloneclusterdf['Strand'][i]
            windowreturn=self.window_sliding(genereads,TSS_start,TSS_end,strand)
            allsortfddict[clusterID]=windowreturn


        print('window sliding Time elapsed',int(time.time()-start_time),'seconds.')


        # allsortfddict={}
        # for geneid,resls in zip(alloneclusterdf['Unnamed: 0'],results):
        #     allsortfddict[geneid]=resls  

        
        ctssOutPath=self.ctss_out_dir+'CTSS_foldchange.pkl'
        with open(ctssOutPath,'wb') as f:
            pickle.dump(allsortfddict,f)

        return allsortfddict


    def pickCTSS(self,ctssls):

        keepCTSS=[]
        for ele in ctssls:
            if (ele[1]>self.minCTSSCount)&(ele[2]>self.minFC):
                keepCTSS.append(ele)
        return keepCTSS





    def produce_CTSS_adata(self):
        ctime=time.time()

 
        #print(self.out_dir)
        self.ctss_out_dir=os.path.join(self.out_dir+'/CTSS/')
        if not os.path.exists(self.ctss_out_dir):
            os.mkdir(self.ctss_out_dir)




        allsortfddict=self._get_CTSS()
        keepdict={}
        for ctssid in allsortfddict.keys():
            keepdict[ctssid]=self.pickCTSS(allsortfddict[ctssid])
        
        #print(keepdict)

        #get the cellID meeting our requirement
        cellIDdict={}
        for i in keepdict.keys():
            
            for j in keepdict[i]:
                geneid=i.split('*')[0]
                newid=i+'#'+str(j[0])+'@'+str(j[1])+'$'+str(j[2])
                cellIDls=[]
                for ele in self.fetchadata[geneid]:
                    if j[0]==ele[0]:
                        cellIDls.append(ele[1])
                cellIDdict[newid]=cellIDls

        #print(len(cellIDdict))


        #create a big matrix including cell ID
        cellidls=list(cellIDdict.values())
        cellidset = set([item for sublist in cellidls for item in sublist])
        ctssfinaldf=pd.DataFrame(index=list(cellidset))



        
        for clusterID in cellIDdict.keys():
            cellID,count=np.unique(cellIDdict[clusterID],return_counts=True)
            CTSSdf=pd.DataFrame({'cell_id':cellID,clusterID:count})
            CTSSdf.set_index('cell_id',inplace=True)
            ctssfinaldf[clusterID]=ctssfinaldf.index.map(CTSSdf[clusterID])


        ctssfinaldf.fillna(0,inplace=True)
        #print(ctssfinaldf)
        ctssadata=ad.AnnData(ctssfinaldf)

        ctssvardf=pd.DataFrame(ctssadata.var.copy())
        ctssvardf.reset_index(inplace=True)
        ctssvardf.columns=['clusterID']
        ctssvardf['gene_id']=ctssvardf['clusterID'].str.split('*',expand=True)[0]
        ctssvardf['CTSS']=ctssvardf['clusterID'].str.split('#',expand=True)[1].str.split('@',expand=True)[0]
        ctssvardf['counts_dropped_UnencodedG']=ctssvardf['clusterID'].str.split('@',expand=True)[1].str.split('$',expand=True)[0]
        ctssvardf['fold_change']=ctssvardf['clusterID'].str.split('$',expand=True)[1]


        ctssvardf=ctssvardf.merge(self.generefdf,on='gene_id')
        ctssvardf.set_index('clusterID',drop=True,inplace=True)
        ctssadata.var=ctssvardf

        ctss_output_h5ad=self.ctss_out_dir+'all_ctss.h5ad'
        ctssadata.write(ctss_output_h5ad)


        twoctssselect=ctssadata.var[ctssadata.var.duplicated('gene_id',keep=False)].index
        twoctssadata=ctssadata[:,twoctssselect]

        sc_output_h5ad=self.ctss_out_dir+'all_ctss_two.h5ad'
        twoctssadata.write(sc_output_h5ad)

        print('produce CTSS h5ad Time elapsed',int(time.time()-ctime),'seconds.')


        return twoctssadata
    



import numpy as np
import pandas as pd
from brie.utils import fetch_reads
import os
import anndata as ad
import multiprocessing 
import pysam
import time
import warnings
import sys
import matplotlib.pyplot as plt


global CACHE_CHROM
global CACHE_SAMFILE
CACHE_CHROM = None
CACHE_SAMFILE = None



def check_pysam_chrom(samFile,chrom=None):
    """
    check samFile is a file name or a pysam object, and if chrom format.
    """
    global CACHE_CHROM
    global CACHE_SAMFILE

    if CACHE_CHROM is not None:
        if (samFile==CACHE_SAMFILE) and (chrom==CACHE_CHROM):
            return CACHE_SAMFILE,CACHE_CHROM

    
    if type(samFile)==str or type(samFile)==np.str_:
        ftype=samFile.split(".")[-1]
        if ftype != "bam" and ftype!="sam" and ftype!="cram":
            print("Error: file type need suffix of bam, sam or cram")
            sys.exit(1)
        if ftype == "cram":
            samFile=pysam.AlignmentFile(samFile,'rc')
        elif ftype=="bam":
            samFile=pysam.AlignmentFile(samFile,'rb')
        else:
            samFile=pysam.AlignmentFile(samFile,'r')

    if chrom is not None:
        if chrom not in samFile.references:
            if chrom.startswith("chr"):
                chrom=chrom.split('chr')[1]
            else:
                chrom="chr"+chrom
        if chrom not in samFile.references:
            print("Can't find reference %s in samFile"%chrom)
            return samFile,None

    CACHE_CHROM=chrom
    CACHE_SAMFILE=samFile
    return samFile,chrom



def _getreads_2(bamfilePath,geneid,generefdf):
    #fetch reads2 in gene 
    samFile, _chrom = check_pysam_chrom(bamfilePath,'chr'+str(generefdf.loc[geneid]['Chromosome']))
    reads = fetch_reads(samFile, _chrom,  generefdf.loc[geneid]['Start'] , generefdf.loc[geneid]['End'],  trimLen_max=100)
    reads2_umi = reads["reads1u"]
#     print(geneid)

    #select according to GX tag and CB (filter according to user owned cell)
    reads2_umi=[r for r in reads2_umi if r.get_tag('GX')==geneid]
    reads2_umi=[r for r in reads2_umi if r.get_tag('CB') in self.cellBarcode]
    
#     print(reads2_umi)
    reads2_position = [r.positions[0] for r in reads2_umi]
    
    return reads2_position

    


from optparse import OptionParser,OptionGroup
from ..version import __version__
import sys
from ..utils.build_ref import get_TSSref,get_generef,get_filter_TSS
from ..utils.get_counts import get_TSS_count
from ..utils.get_ctss import get_CTSS_count
import pyranges as pr
import os
import pandas as pd
import time 


START_TIME = time.time()


def main():
    parser = OptionParser()
    parser.add_option('--gtf','-g',dest='gtf_file',default=None,help='The annotation gtf file for your analysing species.')
    parser.add_option('--cellbarcodeFile','-c',dest='cdrFile',default=None,help='The file include cell barcode which users want to keep in the downstream analysis.')
    parser.add_option('--bam','-b',dest='bam_file',default=None,help='The bam file of aligned from Cellranger or other single cell aligned software.')
    parser.add_option('--outdir','-o',dest='out_dir',default=None,help='The directory for output [default : $bam_file]') #what should be after $
    parser.add_option('--refFasta','-r',dest='refFasta',default=None,help='The directory for reference genome fasta file') #what should be after $
    parser.add_option('--mode','-m',dest='mode',default=None,help='You can select run by finding novel TSS cluster and CTSS within one cluster [TC+CTSS]. \
                        If you just want to detect TSS cluster, you can use [TC] mode. If you just want to detect CTSS, you can use [CTSS] mode which is based on the output of [TC mode]')

   
   
    group0=OptionGroup(parser,"Optional arguments")

    group0.add_option("--minCount",type="int",dest="minCount",default=50,
    help="Minimum UMI counts for TC in all cells [default: 50]")

    group0.add_option('--nproc','-p',type="int",dest='nproc',default=4,
    help='Number of subprocesses [default: 4]')

    group0.add_option('--maxReadCount',type="int",dest='maxReadCount',default=10000,
    help='For each gene, the maxmium read count kept for clustering [default: 10000]')
    
    group0.add_option('--clusterDistance',type="float",dest='clusterDistance',default=300,
    help="The minimum distance between two cluster transcription start site [default: 300]")

    group0.add_option('--InnerDistance',type="float",dest='InnerDistance',default=100,
    help="The resolution of each cluster [default: 100]")

    group0.add_option('--windowSize',type="int",dest='windowSize',default=15,
    help="The width of sliding window [default: 15]")



    group1=OptionGroup(parser,"Optional arguments")

    group1.add_option('--minCTSSCount',type="float",dest='minCTSSCount',default=100,
    help="The minimum UMI counts for each CTSS [default: 100]")

    group1.add_option('--minFC',type="float",dest='minFC',default=6,
    help="The minimum fold change for filtering CTSS [default: 6]")



    parser.add_option_group(group0)
    parser.add_option_group(group1)


    (options, args) = parser.parse_args()

    
    #this means that if users do not input any argument, then direct produce help. then end.
    if len(sys.argv[1:]) == 0:
        print('Welcome to CamoTSS v%s!\n'%(__version__))
        print("use -h or --help for help on argument.")
        sys.exit(1)





    if options.mode is None:
        print("Error: Need --mode to select the mode what you prefer.")
        sys.exit(1)

    
    if (options.mode=='TC') or (options.mode=='TC+CTSS'):
        if options.cdrFile is None:
            print("Error: Need --cdrFile for cell barcode file.")
            sys.exit(1)

        if options.refFasta is None:
            print("Error: Need --refFasta for reference fasta file.")
            sys.exit(1)

        #bam file
        if options.bam_file is None:
            print("Error: Need --bam for aligned file.")
            sys.exit(1)


            #output file 
        if options.out_dir is None:
            print("Warning: no outDir provided, we use $bamfilePath/CamoTSS\n")
            out_dir = os.path.dirname(os.path.abspath(options.bam_file)) + "/CamoTSS"
        else:
            out_dir = options.out_dir
        if not os.path.exists(out_dir):
            os.mkdir(out_dir)


        #gtf file
        if options.gtf_file is None:
            print("Error: Need --gtf for annotation file.")
            sys.exit(1)
        else:
            gr = pr.read_gtf(options.gtf_file)
            grdf = gr.df
            ref_out_dir=str(out_dir)+'/ref_file/'
            if not os.path.exists(ref_out_dir):
                os.mkdir(ref_out_dir)
            tssrefpath=get_TSSref(grdf,ref_out_dir)
            tssdf=pd.read_csv(tssrefpath,delimiter='\t')
            generefpath=get_generef(grdf,tssdf,ref_out_dir)


    elif options.mode=='CTSS':
        if options.out_dir is None:
            print("Error: Need --outdir which includes subdir '/ref_file' and '/count'")
            sys.exit(1)

        out_dir=options.out_dir


    bam_file=options.bam_file
    minCount=options.minCount
    cellBarcodePath=options.cdrFile
    n_proc=options.nproc
    maxReadCount=options.maxReadCount
    clusterDistance=options.clusterDistance
    InnerDistance=options.InnerDistance
    fastqFilePath=options.refFasta
    windowSize=options.windowSize
    minCTSSCount=options.minCTSSCount
    minFC=options.minFC
    




        
    if options.mode == "TC":
        getTSScount=get_TSS_count(generefpath,tssrefpath,bam_file,fastqFilePath,out_dir,cellBarcodePath,n_proc,minCount,maxReadCount,clusterDistance,InnerDistance,windowSize,minCTSSCount,minFC)
        scadata=getTSScount.produce_sclevel()

    elif options.mode=="TC+CTSS":
        # ctss_out_dir=str(options.out_dir)+'/CTSS/'
        # if not os.path.exists(ctss_out_dir):
        #     os.mkdir(ctss_out_dir)
        getTSScount=get_TSS_count(generefpath,tssrefpath,bam_file,fastqFilePath,out_dir,cellBarcodePath,n_proc,minCount,maxReadCount,clusterDistance,InnerDistance,windowSize,minCTSSCount,minFC)
        scadata=getTSScount.produce_sclevel()
        twoctssadata=getTSScount.produce_CTSS_adata()


    elif options.mode=='CTSS':
        getctsscount=get_CTSS_count(out_dir,minCTSSCount,minFC,n_proc,windowSize)   # should create CTSS 
        ctssadata=getctsscount.produce_CTSS_adata()


    else:
        print('Do not have this mode. Please check your spell!')
        run_time = time.time() - START_TIME
        print("[CamoTSS] All done: %d min %.1f sec" %(int(run_time / 60), 
                                                  run_time % 60))


